from pathlib import Path
from typing import Dict, Type, Optional, Any
import importlib
import inspect
import logging
import json
from .base import BaseBalancer


class TechniqueRegistry:
    """Registry for managing balancing techniques from various sources"""

    # Define the modules to check for techniques
    IMBLEARN_MODULES = [
        "imblearn.over_sampling",
        "imblearn.under_sampling",
        "imblearn.combine",
    ]

    def __init__(self):
        self.custom_techniques: Dict[str, Type[BaseBalancer]] = {}
        self._cached_imblearn_techniques: Dict[str, tuple] = {}
        self._discover_imblearn_techniques()
        self._load_custom_techniques()

    def _discover_imblearn_techniques(self) -> None:
        """Dynamically discover all available techniques in imblearn"""
        for module_path in self.IMBLEARN_MODULES:
            try:
                module = importlib.import_module(module_path)
                # Get all classes from the module
                for name, obj in inspect.getmembers(module, inspect.isclass):
                    # Check if it's a sampler (has fit_resample method)
                    if hasattr(obj, "fit_resample"):
                        self._cached_imblearn_techniques[name] = (module_path, obj)
            except ImportError as e:
                logging.warning(f"Could not import {module_path}: {str(e)}")

    def get_technique_class(self, technique_name: str) -> Optional[Type[BaseBalancer]]:
        """Get the technique class by name, handling suffixed variations."""
        # First, check for exact matches in custom techniques
        if technique_name in self.custom_techniques:
            return self.custom_techniques[technique_name]

        # Then check for exact matches in imblearn techniques
        if technique_name in self._cached_imblearn_techniques:
            module_path, technique_class = self._cached_imblearn_techniques[
                technique_name
            ]
            return self._wrap_imblearn_technique(technique_class)

        # If no exact match, extract base name if this is a variation with _ or - suffix
        base_name = None
        for delimiter in ["_", "-"]:
            if delimiter in technique_name:
                parts = technique_name.split(delimiter, 1)
                if len(parts) > 1 and parts[0]:  # Ensure we have a non-empty base name
                    base_name = parts[0]
                    break

        # If we have a valid base name, try to look it up
        if base_name:
            # Check custom techniques for the base name
            if base_name in self.custom_techniques:
                return self.custom_techniques[base_name]

            # Check imblearn techniques for the base name
            if base_name in self._cached_imblearn_techniques:
                module_path, technique_class = self._cached_imblearn_techniques[
                    base_name
                ]
                return self._wrap_imblearn_technique(technique_class)

        # If still not found, try to discover new techniques (in case imblearn was updated)
        self._discover_imblearn_techniques()

        # Try exact match again with freshly discovered techniques
        if technique_name in self._cached_imblearn_techniques:
            module_path, technique_class = self._cached_imblearn_techniques[
                technique_name
            ]
            return self._wrap_imblearn_technique(technique_class)

        # Try base name again with freshly discovered techniques
        if base_name and base_name in self._cached_imblearn_techniques:
            module_path, technique_class = self._cached_imblearn_techniques[base_name]
            return self._wrap_imblearn_technique(technique_class)

        return None

    def get_technique_default_params(self, technique_name: str) -> Dict[str, Any]:
        """
        Extract default parameters from a technique, handling suffixed variations.

        Args:
            technique_name: Name of the technique to extract parameters from

        Returns:
            Dictionary of parameter names and their default values
        """
        # First check for exact match in custom techniques
        if technique_name in self.custom_techniques:
            technique_class = self.custom_techniques[technique_name]
            return self._extract_params_from_class(technique_class)

        # Then check for exact match in imblearn techniques
        if technique_name in self._cached_imblearn_techniques:
            _, technique_class = self._cached_imblearn_techniques[technique_name]
            return self._extract_params_from_class(technique_class)

        # If no exact match, try to find a base name by removing suffix
        base_name = None
        for delimiter in ["_", "-"]:
            if delimiter in technique_name:
                parts = technique_name.split(delimiter, 1)
                if len(parts) > 1 and parts[0]:
                    base_name = parts[0]
                    break

        # If we have a valid base name, look it up
        if base_name:
            # Check custom techniques for the base name
            if base_name in self.custom_techniques:
                technique_class = self.custom_techniques[base_name]
                return self._extract_params_from_class(technique_class)

            # Check imblearn techniques for the base name
            if base_name in self._cached_imblearn_techniques:
                _, technique_class = self._cached_imblearn_techniques[base_name]
                return self._extract_params_from_class(technique_class)

        # If still not found
        logging.warning(
            f"Technique '{technique_name}' not found. Cannot extract parameters."
        )
        return {}

    def _extract_params_from_class(self, cls) -> Dict[str, Any]:
        """
        Extract default parameters from a class's __init__ method.

        Args:
            cls: The class to extract parameters from

        Returns:
            Dictionary of parameter names and their default values
        """
        params = {}

        try:
            # Get the signature of the __init__ method
            sig = inspect.signature(cls.__init__)

            # Process each parameter
            for name, param in sig.parameters.items():
                # Skip 'self' parameter
                if name == "self":
                    continue

                # Get default value if it exists
                if param.default is not inspect.Parameter.empty:
                    # Handle special case for None (JSON uses null)
                    if param.default is None:
                        params[name] = None
                    # Handle other types that can be serialised to JSON
                    elif isinstance(param.default, (int, float, str, bool, list, dict)):
                        params[name] = param.default
                    else:
                        # Convert non-JSON-serialisable defaults to string representation
                        params[name] = str(param.default)
                else:
                    # For parameters without defaults, use None
                    params[name] = None

        except Exception as e:
            logging.warning(f"Error extracting parameters from {cls.__name__}: {e}")

        return params

    def list_available_techniques(self) -> Dict[str, list]:
        """List all available techniques grouped by source"""
        # Rediscover techniques in case new ones were added
        self._discover_imblearn_techniques()

        return {
            "custom": list(self.custom_techniques.keys()),
            "imblearn": list(self._cached_imblearn_techniques.keys()),
        }

    def register_custom_technique(
        self, name: str, technique_class: Type[BaseBalancer]
    ) -> None:
        """
        Register a custom balancing technique.

        Args:
            name: Name of the technique
            technique_class: Class implementing the balancing technique

        Raises:
            TypeError: If technique_class is None or doesn't inherit from BaseBalancer
            ValueError: If name is empty or not a string
        """
        # Error handling
        if not isinstance(name, str) or not name.strip():
            raise ValueError("Technique name must be a non-empty string")

        if technique_class is None:
            raise TypeError("Technique class cannot be None")

        if not isinstance(technique_class, type) or not issubclass(
            technique_class, BaseBalancer
        ):
            raise TypeError("Technique class must inherit from BaseBalancer")
        self.custom_techniques[name] = technique_class

    def _load_custom_techniques(self) -> None:
        """Load registered custom techniques from the custom techniques directory."""
        custom_dir = Path.home() / ".balancr" / "custom_techniques"
        if not custom_dir.exists():
            return

        metadata_file = custom_dir / "techniques_metadata.json"
        if not metadata_file.exists():
            return

        try:
            with open(metadata_file, "r") as f:
                metadata = json.load(f)

            for technique_name, info in metadata.items():
                file_path = Path(info["file"])
                class_name = info["class_name"]

                if not file_path.exists():
                    logging.warning(f"Custom technique file not found: {file_path}")
                    continue

                try:
                    # Import the module dynamically
                    module_name = file_path.stem
                    spec = importlib.util.spec_from_file_location(
                        module_name, file_path
                    )
                    if spec is None or spec.loader is None:
                        logging.warning(f"Could not load module from {file_path}")
                        continue

                    module = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(module)

                    # Find the specific class
                    technique_class = None
                    for name, obj in inspect.getmembers(module, inspect.isclass):
                        if name == class_name and issubclass(obj, BaseBalancer):
                            technique_class = obj
                            break

                    if technique_class:
                        self.custom_techniques[technique_name] = technique_class
                        logging.debug(f"Loaded custom technique: {technique_name}")
                    else:
                        logging.warning(f"Class {class_name} not found in {file_path}")

                except Exception as e:
                    logging.warning(
                        f"Error loading custom technique {technique_name}: {e}"
                    )

        except Exception as e:
            logging.warning(f"Error loading custom techniques metadata: {e}")

    def _wrap_imblearn_technique(self, technique_class: type) -> Type[BaseBalancer]:
        """Wrap imblearn technique to conform to our BaseBalancer interface"""

        class WrappedTechnique(BaseBalancer):
            def __init__(self, **kwargs):
                super().__init__()
                self.technique = technique_class(**kwargs)

            def balance(self, X, y):
                return self.technique.fit_resample(X, y)

        return WrappedTechnique



"""Metrics for evaluating imbalanced classification performance."""

import logging
import time
from typing import Dict
import numpy as np
import pandas as pd
from sklearn.base import clone
from sklearn.metrics import (
    accuracy_score,
    precision_score,
    recall_score,
    f1_score,
    roc_auc_score,
    average_precision_score,
    confusion_matrix,
)
from sklearn.model_selection import StratifiedKFold, learning_curve


def format_time(seconds):
    """Format time in seconds to minutes and seconds"""
    minutes = int(seconds // 60)
    remaining_seconds = seconds % 60
    return f"{minutes}mins, {remaining_seconds:.2f}secs"


def get_metrics(
    classifier,
    X_test: np.ndarray,
    y_test: np.ndarray,
) -> Dict[str, float]:
    """
    Calculate metrics specifically suited for imbalanced classification.
    Works with both binary and multiclass problems.

    Args:
        classifier: Pre-fitted classifier instance to evaluate
        X_test: Test features
        y_test: Test labels

    Returns:
        Dictionary containing various metric scores
    """
    # Get predictions
    y_pred = classifier.predict(X_test)

    metrics = {}

    # For ROC AUC, we need probability predictions
    if hasattr(classifier, "predict_proba"):
        try:
            y_pred_proba = classifier.predict_proba(X_test)
            # For multiclass problems, we'll use different methods below
        except (AttributeError, IndexError):
            # Log warning and set probability metrics to NaN
            logging.warning(
                f"Failed to get probability predictions from classifier {classifier.__class__.__name__}. "
                "ROC-AUC and average_precision metrics will be set to NaN."
            )
            metrics["roc_auc"] = float("nan")
            metrics["average_precision"] = float("nan")
            # Skip the rest of probability-based calculations
            y_pred_proba = None
    else:
        # Log warning and set probability metrics to NaN
        logging.warning(
            f"Classifier {classifier.__class__.__name__} does not support predict_proba. "
            "ROC-AUC and average_precision metrics will be set to NaN."
        )
        metrics["roc_auc"] = float("nan")
        metrics["average_precision"] = float("nan")
        # Skip the rest of probability-based calculations
        y_pred_proba = None

    # Calculate confusion matrix
    cm = confusion_matrix(y_test, y_pred)
    n_classes = len(np.unique(y_test))

    # Calculate metrics based on number of classes
    if n_classes == 2:  # Binary classification
        # Unpack binary confusion matrix
        tn, fp, fn, tp = cm.ravel()

        # Calculate binary metrics
        specificity = tn / (tn + fp) if (tn + fp) > 0 else 0
        g_mean = np.sqrt(recall_score(y_test, y_pred) * specificity)

        # Create metrics dictionary for binary case
        metrics = {
            "accuracy": accuracy_score(y_test, y_pred),
            "precision": precision_score(y_test, y_pred),
            "recall": recall_score(y_test, y_pred),
            "specificity": specificity,
            "f1": f1_score(y_test, y_pred),
            "g_mean": g_mean,
        }

        # Add probability-based metrics if possible
        try:
            # For binary classification, we need the probability of the positive class
            if (
                isinstance(y_pred_proba, np.ndarray)
                and y_pred_proba.ndim > 1
                and y_pred_proba.shape[1] > 1
            ):
                y_pred_proba_pos = y_pred_proba[:, 1]
            else:
                y_pred_proba_pos = y_pred_proba

            metrics["roc_auc"] = roc_auc_score(y_test, y_pred_proba_pos)
            metrics["average_precision"] = average_precision_score(
                y_test, y_pred_proba_pos
            )
        except (ValueError, TypeError):
            # Skip these metrics if they can't be calculated
            metrics["roc_auc"] = float("nan")
            metrics["average_precision"] = float("nan")

    else:  # Multiclass classification
        # Calculate per-class specificity and g-mean
        specificities = []
        recalls = []

        for i in range(n_classes):
            # For each class, treat it as positive and all others as negative
            y_true_binary = (y_test == i).astype(int)
            y_pred_binary = (y_pred == i).astype(int)

            # Calculate per-class CM values
            cm_i = confusion_matrix(y_true_binary, y_pred_binary)
            if cm_i.shape[0] < 2:  # Handle edge case
                specificities.append(0)
                recalls.append(0)
                continue

            tn_i, fp_i, fn_i, tp_i = cm_i.ravel()

            # Calculate specificity and recall for this class
            spec_i = tn_i / (tn_i + fp_i) if (tn_i + fp_i) > 0 else 0
            rec_i = tp_i / (tp_i + fn_i) if (tp_i + fn_i) > 0 else 0

            specificities.append(spec_i)
            recalls.append(rec_i)

        # Calculate macro-averaged metrics
        macro_specificity = np.mean(specificities)
        macro_recall = np.mean(recalls)
        g_mean = np.sqrt(macro_recall * macro_specificity)

        # Create metrics dictionary for multiclass case
        metrics = {
            "accuracy": accuracy_score(y_test, y_pred),
            "precision": precision_score(y_test, y_pred, average="macro"),
            "recall": recall_score(y_test, y_pred, average="macro"),
            "specificity": macro_specificity,
            "f1": f1_score(y_test, y_pred, average="macro"),
            "g_mean": g_mean,
        }

        # Add multiclass ROC AUC if possible
        try:
            # For multiclass, use roc_auc_score with multi_class parameter
            if hasattr(classifier, "predict_proba"):
                metrics["roc_auc"] = roc_auc_score(
                    y_test, y_pred_proba, multi_class="ovr", average="macro"
                )

                # Calculate average precision for multiclass
                metrics["average_precision"] = precision_score(
                    y_test, y_pred, average="macro"
                )
            else:
                metrics["roc_auc"] = float("nan")
                metrics["average_precision"] = float("nan")
        except (ValueError, TypeError):
            # Skip these metrics if they can't be calculated
            metrics["roc_auc"] = float("nan")
            metrics["average_precision"] = float("nan")

    return metrics


def get_cv_scores(
    classifier,
    X_balanced: np.ndarray,
    y_balanced: np.ndarray,
    n_folds: int = 5,
) -> Dict[str, float]:
    """
    Perform cross-validation and return average scores.
    Automatically handles multiclass data by using macro averaging.

    Args:
        classifier: Classifier instance to evaluate
        X_balanced: Balanced feature matrix
        y_balanced: Balanced target vector
        n_folds: Number of cross-validation folds

    Returns:
        Dictionary containing average metric scores
    """
    import logging
    from sklearn.model_selection import cross_val_score, cross_val_predict
    from sklearn.metrics import roc_auc_score, confusion_matrix
    import numpy as np

    # Determine if we're dealing with multiclass data
    unique_classes = np.unique(y_balanced)
    is_multiclass = len(unique_classes) > 2

    # Initialise metrics dictionary
    metrics = {}

    # Use scikit-learn's cross_val_score for standard metrics
    # Accuracy doesn't need special handling for multiclass
    scores = cross_val_score(
        classifier, X_balanced, y_balanced, cv=n_folds, scoring="accuracy"
    )
    metrics["cv_accuracy_mean"] = scores.mean()
    metrics["cv_accuracy_std"] = scores.std()

    # For metrics that need proper multiclass handling
    for metric in ["precision", "recall", "f1"]:
        # Use macro averaging for multiclass problems
        if is_multiclass:
            scoring = f"{metric}_macro"
        else:
            scoring = metric

        scores = cross_val_score(
            classifier, X_balanced, y_balanced, cv=n_folds, scoring=scoring
        )
        metrics[f"cv_{metric}_mean"] = scores.mean()
        metrics[f"cv_{metric}_std"] = scores.std()

    # For ROC-AUC and G-mean, we need the predictions
    try:
        # Get cross-validated predictions
        y_pred = cross_val_predict(classifier, X_balanced, y_balanced, cv=n_folds)

        # Calculate G-mean
        if is_multiclass:
            # Calculate per-class specificities and recalls for each class
            specificities = []
            recalls = []

            for i in unique_classes:
                # For each class, treat it as positive and all others as negative
                y_true_binary = (y_balanced == i).astype(int)
                y_pred_binary = (y_pred == i).astype(int)

                # Calculate confusion matrix
                cm = confusion_matrix(y_true_binary, y_pred_binary)

                # Ensure the confusion matrix has the right shape
                if cm.shape == (2, 2):
                    tn, fp, fn, tp = cm.ravel()

                    # Calculate specificity and recall
                    specificity = tn / (tn + fp) if (tn + fp) > 0 else 0
                    recall = tp / (tp + fn) if (tp + fn) > 0 else 0

                    specificities.append(specificity)
                    recalls.append(recall)

            # Calculate macro-averaged G-mean
            if specificities and recalls:  # Check if lists are not empty
                macro_specificity = np.mean(specificities)
                macro_recall = np.mean(recalls)
                g_mean = np.sqrt(macro_specificity * macro_recall)

                metrics["cv_g_mean_mean"] = g_mean
                metrics["cv_g_mean_std"] = (
                    0.0  # Cannot calculate std from a single value
                )
            else:
                logging.warning(
                    f"Could not calculate G-mean for classifier {classifier.__class__.__name__}. "
                    "Some classes may not have been predicted correctly."
                )
                # We don't set the metrics here, letting the absence indicate an issue
        else:
            # Binary case
            cm = confusion_matrix(y_balanced, y_pred)

            if cm.shape == (2, 2):
                tn, fp, fn, tp = cm.ravel()

                specificity = tn / (tn + fp) if (tn + fp) > 0 else 0
                sensitivity = tp / (tp + fn) if (tp + fn) > 0 else 0

                g_mean = np.sqrt(specificity * sensitivity)
                metrics["cv_g_mean_mean"] = g_mean
            else:
                logging.warning(
                    f"Could not calculate G-mean for classifier {classifier.__class__.__name__}. "
                    "Unexpected confusion matrix shape."
                )
                # Metrics not set
    except Exception as e:
        logging.warning(
            f"Could not calculate G-mean for classifier {classifier.__class__.__name__}. "
            f"Error: {str(e)}"
        )
        # Metrics not set

    # ROC-AUC calculation - separate try block to ensure G-mean calculation happens even if ROC-AUC fails
    try:
        if hasattr(classifier, "predict_proba"):
            # Get probability predictions
            if is_multiclass:
                y_proba = cross_val_predict(
                    classifier,
                    X_balanced,
                    y_balanced,
                    cv=n_folds,
                    method="predict_proba",
                )
                roc_auc = roc_auc_score(
                    y_balanced, y_proba, multi_class="ovr", average="macro"
                )
            else:
                y_proba = cross_val_predict(
                    classifier,
                    X_balanced,
                    y_balanced,
                    cv=n_folds,
                    method="predict_proba",
                )
                # Use second column for positive class probability
                if y_proba.shape[1] > 1:
                    roc_auc = roc_auc_score(y_balanced, y_proba[:, 1])
                else:
                    roc_auc = roc_auc_score(y_balanced, y_proba)

            metrics["cv_roc_auc_mean"] = roc_auc
        else:
            logging.warning(
                f"Classifier {classifier.__class__.__name__} does not support predict_proba. "
                "ROC-AUC cannot be calculated."
            )
            # Metrics not set
    except Exception as e:
        logging.warning(
            f"Could not calculate ROC-AUC for classifier {classifier.__class__.__name__}. "
            f"Error: {str(e)}"
        )
        # Metrics not set

    return metrics


def get_learning_curve_data(
    classifier,
    X: np.ndarray,
    y: np.ndarray,
    train_sizes: np.ndarray = np.linspace(0.1, 1.0, 10),
    n_folds: int = 5,
) -> Dict[str, np.ndarray]:
    """
    Compute data for plotting learning curves.

    Args:
        classifier: Classifier instance to evaluate
        X: Feature matrix
        y: Target vector
        train_sizes: Relative or absolute sizes of the training dataset
        n_folds: Number of cross-validation folds

    Returns:
        Dictionary containing training sizes, training scores, and validation scores
    """
    train_sizes_abs, train_scores, val_scores = learning_curve(
        estimator=classifier,
        X=X,
        y=y,
        train_sizes=train_sizes,
        cv=n_folds,
        scoring="accuracy",  # Default metric is accuracy
        shuffle=True,
    )

    return {
        "train_sizes": train_sizes_abs,
        "train_scores": train_scores,
        "val_scores": val_scores,
    }


def get_learning_curve_data_multiple_techniques(
    classifier_name: str,
    classifier,
    techniques_data: Dict[str, Dict[str, np.ndarray]],
    train_sizes: np.ndarray = np.linspace(0.1, 1.0, 10),
    n_folds: int = 5,
) -> Dict[str, Dict[str, np.ndarray]]:
    """
    Compute data for plotting learning curves for multiple techniques.

    Args:
        classifier: Classifier instance to evaluate
        techniques_data: A dictionary where keys are technique names and values are dictionaries
                          containing 'X_balanced' and 'y_balanced' for each technique
        train_sizes: Relative or absolute sizes of the training dataset
        n_folds: Number of cross-validation folds

    Returns:
        Dictionary containing training sizes, training scores, and validation scores for each technique
    """
    learning_curve_data = {}

    # Loop through each technique's data
    for technique_name, data in techniques_data.items():
        X_balanced = data["X_balanced"]
        y_balanced = data["y_balanced"]

        start_time = time.time()
        logging.info(
            f"Generating learning curve for {classifier_name} trained on data "
            f"balanced by {technique_name}, against balanced data..."
        )
        train_sizes_abs, train_scores, val_scores = learning_curve(
            estimator=classifier,
            X=X_balanced,
            y=y_balanced,
            train_sizes=train_sizes,
            cv=n_folds,
            scoring="accuracy",  # Default metric is accuracy
            shuffle=True,
        )
        curve_generating_time = time.time() - start_time
        logging.info(
            f"Generated learning curve for {classifier_name} trained on data "
            f"balanced by {technique_name}, against balanced data"
            f"(Time Taken: {format_time(curve_generating_time)})"
        )

        learning_curve_data[technique_name] = {
            "train_sizes": train_sizes_abs,
            "train_scores": train_scores,
            "val_scores": val_scores,
        }

    return learning_curve_data


def get_learning_curve_data_against_imbalanced_multiple_techniques(
    classifier_name: str,
    classifier,
    techniques_data: Dict[str, Dict[str, np.ndarray]],
    X_test,
    y_test,
    train_sizes: np.ndarray = np.linspace(0.1, 1.0, 10),
    n_folds: int = 5,
) -> Dict[str, Dict[str, np.ndarray]]:
    """
    Custom learning curve function that trains on balanced data
    and evaluates on original imbalanced test set.

    Returns training and validation scores at each learning curve point.
    Validation scores are computed on X_test / y_test.
    """

    def safe_index(X, indices):
        return X.iloc[indices] if isinstance(X, (pd.DataFrame, pd.Series)) else X[indices]

    learning_curve_data = {}

    for technique_name, data in techniques_data.items():
        X_balanced = data["X_balanced"]
        y_balanced = data["y_balanced"]

        n_samples = X_balanced.shape[0]
        train_sizes_abs = (train_sizes * n_samples).astype(int)

        train_scores = []
        val_scores = []

        start_time = time.time()
        logging.info(
            f"Generating learning curve for {classifier_name} trained on data "
            f"balanced by {technique_name}, against original test data..."
        )

        for train_size in train_sizes_abs:
            X_subset = safe_index(X_balanced, np.arange(train_size))
            y_subset = safe_index(y_balanced, np.arange(train_size))

            fold_train_scores = []
            fold_val_scores = []

            kf = StratifiedKFold(n_splits=n_folds, shuffle=True, random_state=42)

            for train_idx, _ in kf.split(X_subset, y_subset):
                X_fold_train = safe_index(X_subset, train_idx)
                y_fold_train = safe_index(y_subset, train_idx)

                clf = clone(classifier)
                clf.fit(X_fold_train, y_fold_train)

                train_acc = clf.score(X_fold_train, y_fold_train)
                val_acc = clf.score(X_test, y_test)

                fold_train_scores.append(train_acc)
                fold_val_scores.append(val_acc)

            train_scores.append(fold_train_scores)
            val_scores.append(fold_val_scores)

        curve_generating_time = time.time() - start_time
        logging.info(
            f"Generated learning curve for {classifier_name} trained on data "
            f"balanced by {technique_name}, against original test data "
            f"(Time Taken: {format_time(curve_generating_time)})"
        )

        learning_curve_data[technique_name] = {
            "train_sizes": train_sizes_abs,
            "train_scores": np.array(train_scores),
            "val_scores": np.array(val_scores),
        }

    return learning_curve_data



import pytest
import numpy as np
import pandas as pd
from unittest.mock import patch, MagicMock
from balancr import BalancingFramework, BaseBalancer


# Mock classes and fixtures
class MockTechnique(BaseBalancer):
    def balance(self, X, y):
        # Return the same data for testing
        return X, y


@pytest.fixture
def sample_data():
    """Create sample data for testing"""
    X = np.array([[1, 2], [3, 4], [5, 6], [7, 8]])
    y = np.array([0, 0, 1, 1])
    return X, y


@pytest.fixture
def sample_dataframe(sample_data):
    """Create a sample DataFrame"""
    X, y = sample_data
    df = pd.DataFrame(X, columns=["feature_1", "feature_2"])
    df["target"] = y
    return df


@pytest.fixture
def framework():
    """Create a new framework instance for each test"""
    return BalancingFramework()


@pytest.fixture
def mock_classifier():
    """Create a mock classifier that returns predictable results"""
    mock_clf = MagicMock()
    mock_clf.predict.return_value = np.array([0, 1, 0, 1])
    mock_clf.predict_proba.return_value = np.array(
        [[0.8, 0.2], [0.3, 0.7], [0.9, 0.1], [0.2, 0.8]]
    )

    return mock_clf


@pytest.fixture
def framework_with_balanced_data(mock_classifier):
    """Create a framework instance with mock balanced datasets"""
    framework = BalancingFramework()

    # Test data
    framework.X_test = np.array([[1, 2], [3, 4], [5, 6], [7, 8]])
    framework.y_test = np.array([0, 1, 0, 1])

    # Mock balanced datasets
    framework.current_balanced_datasets = {
        "SMOTE": {
            "X_balanced": np.array([[1, 2], [3, 4], [5, 6], [7, 8]]),
            "y_balanced": np.array([0, 1, 0, 1]),
        },
        "RandomUnderSampler": {
            "X_balanced": np.array([[1, 2], [7, 8]]),
            "y_balanced": np.array([0, 1]),
        },
    }

    # Mock classifier registry
    framework.classifier_registry = MagicMock()

    # Mock classifier class that returns our mock classifier
    mock_classifier_class = MagicMock()
    mock_classifier_class.return_value = mock_classifier

    framework.classifier_registry.get_classifier_class.return_value = (
        mock_classifier_class
    )

    return framework


@pytest.fixture
def mock_registry(monkeypatch):
    """Create a mock registry with test techniques"""
    mock = MagicMock()
    mock.list_available_techniques.return_value = {
        "custom": ["CustomTechnique"],
        "imblearn": ["SMOTE", "RandomUnderSampler"],
    }
    mock.get_technique_class.return_value = MockTechnique
    monkeypatch.setattr(
        "balancr.technique_registry.TechniqueRegistry", lambda: mock
    )
    return mock


def test_train_classifiers_no_balanced_data():
    """Test train_classifiers with no balanced datasets"""
    framework = BalancingFramework()

    with pytest.raises(ValueError, match="No balanced datasets available"):
        framework.train_classifiers()


def test_train_classifiers_no_test_data(framework_with_balanced_data):
    """Test train_classifiers with no test data"""
    framework = framework_with_balanced_data
    framework.X_test = None
    framework.y_test = None

    with pytest.raises(ValueError, match="Test data not found"):
        framework.train_classifiers()


@patch("balancr.imbalance_analyser.get_metrics")
def test_train_classifiers_default_classifier(
    mock_get_metrics, framework_with_balanced_data
):
    """Test train_classifiers with default classifier"""
    framework = framework_with_balanced_data

    # Mock the get_metrics function
    mock_get_metrics.return_value = {
        "accuracy": 0.75,
        "precision": 0.8,
        "recall": 0.7,
        "f1": 0.75,
    }

    # Call the function with default parameters
    results = framework.train_classifiers()

    # Check that results have the expected structure
    assert "RandomForestClassifier" in results
    assert "SMOTE" in results["RandomForestClassifier"]
    assert "RandomUnderSampler" in results["RandomForestClassifier"]
    assert "standard_metrics" in results["RandomForestClassifier"]["SMOTE"]

    # Check that the classifier registry was called correctly
    framework.classifier_registry.get_classifier_class.assert_called_with(
        "RandomForestClassifier"
    )

    # Check that get_metrics was called for each technique
    assert mock_get_metrics.call_count == 2


@patch("balancr.imbalance_analyser.get_metrics")
@patch("balancr.imbalance_analyser.get_cv_scores")
def test_train_classifiers_with_cv(
    mock_get_cv_scores, mock_get_metrics, framework_with_balanced_data
):
    """Test train_classifiers with cross-validation enabled"""
    framework = framework_with_balanced_data

    # Mock metric functions
    mock_get_metrics.return_value = {
        "accuracy": 0.75,
        "precision": 0.8,
        "recall": 0.7,
        "f1": 0.75,
    }

    mock_get_cv_scores.return_value = {
        "cv_accuracy_mean": 0.73,
        "cv_accuracy_std": 0.05,
        "cv_precision_mean": 0.78,
        "cv_precision_std": 0.07,
    }

    # Call function with CV enabled
    results = framework.train_classifiers(enable_cv=True, cv_folds=3)

    # Check that CV metrics are included
    assert "cv_metrics" in results["RandomForestClassifier"]["SMOTE"]
    assert "cv_metrics" in results["RandomForestClassifier"]["RandomUnderSampler"]

    # Check that get_cv_scores was called with the right parameters
    mock_get_cv_scores.assert_called_with(
        framework.classifier_registry.get_classifier_class().return_value,
        framework.current_balanced_datasets["RandomUnderSampler"]["X_balanced"],
        framework.current_balanced_datasets["RandomUnderSampler"]["y_balanced"],
        n_folds=3,
    )


@patch("balancr.imbalance_analyser.get_metrics")
def test_train_classifiers_custom_classifiers(
    mock_get_metrics, framework_with_balanced_data
):
    """Test train_classifiers with custom classifier configurations"""
    framework = framework_with_balanced_data

    # Create two separate mock classifiers
    mock_rf = MagicMock()
    mock_lr = MagicMock()

    # Update the side_effect to use a function that will return different mocks
    def get_classifier_by_name(name):
        if name == "RandomForestClassifier":
            return lambda **kwargs: mock_rf
        elif name == "LogisticRegression":
            return lambda **kwargs: mock_lr
        else:
            return None

    framework.classifier_registry.get_classifier_class.side_effect = (
        get_classifier_by_name
    )

    # Mock get_metrics
    mock_get_metrics.return_value = {
        "accuracy": 0.75,
        "precision": 0.8,
        "recall": 0.7,
        "f1": 0.75,
    }

    # Define custom classifiers
    classifier_configs = {
        "RandomForestClassifier": {"n_estimators": 100, "random_state": 42},
        "LogisticRegression": {"C": 1.0, "random_state": 42},
        "NonExistentClassifier": {"param": "value"},
    }

    # Call the function with custom classifiers
    results = framework.train_classifiers(classifier_configs=classifier_configs)

    assert "RandomForestClassifier" in results
    assert "LogisticRegression" in results
    assert "NonExistentClassifier" not in results

    # Verify that both mocks were used
    assert mock_rf.fit.called
    assert mock_lr.fit.called

    # Check that both SMOTE and RandomUnderSampler were processed for each classifier
    techniques = results["RandomForestClassifier"].keys()
    assert "SMOTE" in techniques
    assert "RandomUnderSampler" in techniques

    techniques = results["LogisticRegression"].keys()
    assert "SMOTE" in techniques
    assert "RandomUnderSampler" in techniques


@patch("balancr.imbalance_analyser.get_metrics")
@patch("logging.error")
def test_train_classifiers_with_error(
    mock_logging_error, mock_get_metrics, framework_with_balanced_data
):
    """Test train_classifiers handling of exceptions during training"""
    framework = framework_with_balanced_data

    # Make the classifier raise an exception for one technique
    framework.classifier_registry.get_classifier_class().return_value.fit.side_effect = [
        Exception("Training error"),
        None,
    ]

    # Mock get_metrics for the successful technique
    mock_get_metrics.return_value = {
        "accuracy": 0.75,
        "precision": 0.8,
        "recall": 0.7,
        "f1": 0.75,
    }

    results = framework.train_classifiers()

    # Check that only the successful technique is in the results
    assert "RandomForestClassifier" in results
    assert len(results["RandomForestClassifier"]) == 1
    assert "RandomUnderSampler" in results["RandomForestClassifier"]
    assert "SMOTE" not in results["RandomForestClassifier"]

    # Check that the error was logged
    mock_logging_error.assert_called_once()
    assert "Training error" in mock_logging_error.call_args[0][0]


def test_train_classifiers_no_classifiers_found(framework_with_balanced_data):
    """Test train_classifiers when no valid classifiers are found"""
    framework = framework_with_balanced_data

    # Make the classifier registry return None for all classifiers
    framework.classifier_registry.get_classifier_class.return_value = None

    results = framework.train_classifiers()

    assert results == {}


def test_framework_initialization(framework):
    """Test framework initialization"""
    assert framework.X is None
    assert framework.y is None
    assert isinstance(framework.results, dict)
    assert isinstance(framework.current_data_info, dict)
    assert isinstance(framework.current_balanced_datasets, dict)


@patch("pandas.read_csv")
def test_load_data_csv(mock_read_csv, framework, sample_dataframe):
    """Test loading data from CSV"""
    mock_read_csv.return_value = sample_dataframe

    framework.load_data(
        file_path="test.csv",
        target_column="target",
        feature_columns=["feature_1", "feature_2"],
    )

    assert framework.X is not None
    assert framework.y is not None
    assert framework.current_data_info["file_path"] == "test.csv"
    assert framework.current_data_info["target_column"] == "target"


@patch("pandas.read_excel")
def test_load_data_excel(mock_read_excel, framework, sample_dataframe):
    """Test loading data from Excel"""
    mock_read_excel.return_value = sample_dataframe

    framework.load_data(
        file_path="test.xlsx",
        target_column="target",
        feature_columns=["feature_1", "feature_2"],
    )

    assert framework.X is not None
    assert framework.y is not None
    assert framework.current_data_info["file_path"] == "test.xlsx"
    assert framework.current_data_info["target_column"] == "target"


def test_load_data_invalid_file(framework):
    """Test loading data from unsupported file format"""
    with pytest.raises(ValueError):
        framework.load_data(file_path="test.txt", target_column="target")


def test_inspect_class_distribution(framework, sample_data):
    """Test class distribution inspection"""
    X, y = sample_data
    framework.X = X
    framework.y = y

    distribution = framework.inspect_class_distribution(display=False)
    assert isinstance(distribution, dict)
    assert set(distribution.keys()) == {0, 1}
    assert distribution[0] == 2  # Count of class 0
    assert distribution[1] == 2  # Count of class 1


def test_preprocess_data_no_data(framework):
    """Test preprocessing without loading data first"""
    with pytest.raises(ValueError):
        framework.preprocess_data()


def test_preprocess_data(framework, sample_data):
    """Test data preprocessing"""
    X, y = sample_data
    framework.X = X
    framework.y = y

    framework.preprocess_data()
    assert framework.X is not None
    assert framework.y is not None
    # Check if data was scaled (mean should be close to 0)
    assert (framework.X.mean().abs() < 1e-10).all()


def test_compare_techniques_no_data(framework):
    """Test comparing techniques without loading data"""
    with pytest.raises(ValueError):
        framework.apply_balancing_techniques(["SMOTE"])


def test_compare_techniques(framework):
    """Test technique comparison with sufficient data for SMOTE"""
    # Create larger dataset that works with SMOTE. Could not mock behaviour properly
    np.random.seed(42)
    X = np.random.rand(20, 5)
    y = np.concatenate([np.zeros(10), np.ones(10)])

    framework.X = X
    framework.y = y

    results = framework.apply_balancing_techniques(
        technique_names=["SMOTE", "RandomUnderSampler"]
    )

    assert isinstance(results, dict)
    assert len(results) == 2
    assert "SMOTE" in results
    assert "RandomUnderSampler" in results


def test_apply_balancing_techniques_invalid_technique(framework_with_balanced_data):
    """Test apply_balancing_techniques with an invalid technique name"""
    framework = framework_with_balanced_data

    # Test data
    framework.X = np.array([[1, 2], [3, 4], [5, 6], [7, 8]])
    framework.y = np.array([0, 1, 0, 1])

    # Configure technique registry to return None for an invalid technique
    framework.technique_registry = MagicMock()
    framework.technique_registry.get_technique_class.return_value = None
    framework.list_available_techniques = MagicMock(
        return_value={"custom": [], "imblearn": ["SMOTE", "RandomUnderSampler"]}
    )

    # Test with invalid technique name
    with pytest.raises(ValueError) as excinfo:
        framework.apply_balancing_techniques(["InvalidTechnique"])

    # Verify the error message
    error_message = str(excinfo.value)
    assert "Technique 'InvalidTechnique' not found" in error_message
    assert "Available techniques" in error_message

    # Verify list_available_techniques was called
    framework.list_available_techniques.assert_called_once()


def test_save_results_no_results(framework):
    """Test saving results without running comparison first"""
    with pytest.raises(ValueError):
        framework.save_results("results.csv")


def test_save_results(framework, tmp_path):
    """Test saving results to file"""
    framework.results = {
        "SMOTE": {
            "accuracy": 0.8,
            "precision": 0.7,
            "recall": 0.9,
            "f1": 0.8,
            "roc_auc": 0.85,
        },
        "RandomUnderSampler": {
            "accuracy": 0.75,
            "precision": 0.8,
            "recall": 0.6,
            "f1": 0.7,
            "roc_auc": 0.82,
        },
    }

    # Test CSV saving
    csv_path = tmp_path / "results.csv"
    framework.save_results(csv_path, file_type="csv", include_plots=False)
    assert csv_path.exists()

    # Test JSON saving
    json_path = tmp_path / "results.json"
    framework.save_results(json_path, file_type="json", include_plots=False)
    assert json_path.exists()


def test_save_results_invalid_type(framework):
    """Test saving results with invalid file type"""
    framework.results = {"test": {"metric": 0.5}}
    with pytest.raises(ValueError):
        framework.save_results("results.txt", file_type="txt")


def test_generate_balanced_data_no_data(framework):
    """Test generating balanced data without running comparison first"""
    with pytest.raises(ValueError):
        framework.generate_balanced_data("output/")


def test_generate_balanced_data(framework, sample_data, tmp_path):
    """Test generating balanced datasets"""
    X, y = sample_data
    framework.X = X
    framework.y = y
    framework.current_data_info = {
        "feature_columns": ["feature_1", "feature_2"],
        "target_column": "target",
    }
    framework.current_balanced_datasets = {"SMOTE": {"X_balanced": X, "y_balanced": y}}

    output_dir = tmp_path / "balanced"
    framework.generate_balanced_data(str(output_dir))

    assert (output_dir / "balanced_SMOTE.csv").exists()


def test_compare_balanced_class_distributions(framework, sample_data, tmp_path):
    """Test comparing class distributions"""
    X, y = sample_data
    framework.current_balanced_datasets = {
        "SMOTE": {"X_balanced": X, "y_balanced": y},
        "RandomUnderSampler": {"X_balanced": X, "y_balanced": y},
    }

    balanced_class_path = tmp_path / "balanced_class_distributions.png"
    framework.compare_balanced_class_distributions(save_path=balanced_class_path)
    assert balanced_class_path.exists()


def test_compare_balanced_class_distributions_no_data(framework):
    """Test comparing class distributions without data"""
    with pytest.raises(ValueError):
        framework.compare_balanced_class_distributions()


def test_generate_learning_curves(framework, tmp_path):
    """Test learning curve generation with sufficient data"""
    # Larger dataset
    np.random.seed(42)
    X = np.random.rand(20, 5)
    y = np.concatenate([np.zeros(10), np.ones(10)])

    framework.current_balanced_datasets = {"SMOTE": {"X_balanced": X, "y_balanced": y}}

    learning_curve_path = tmp_path / "learning_curve.png"
    framework.generate_learning_curves(
        classifier_name="RandomForestClassifier",
        save_path=learning_curve_path,
        learning_curve_type="Balanced Datasets"
    )
    assert learning_curve_path.exists()


def test_generate_learning_curves_no_data(framework):
    """Test learning curve generation without data"""
    with pytest.raises(ValueError):
        framework.generate_learning_curves(
            classifier_name="RandomForestClassifier",
            learning_curve_type="Balanced Datasets"
        )


def test_handle_quality_issues(framework):
    """Test handling of data quality issues"""
    quality_report = {
        "missing_values": [(0, 1)],  # row 0, column 1 missing
        "constant_features": [("feature1", 0)],
        "feature_correlations": [("feature1", "feature2", 0.96)],
    }

    # Should not raise any exceptions
    framework._handle_quality_issues(quality_report)



"""Tests for data-related commands in the balancr CLI."""

import pytest
from unittest.mock import patch, MagicMock

from balancr.cli import commands


@pytest.fixture
def mock_config_path(tmp_path):
    """Create a temporary config file path for testing."""
    return tmp_path / "test_config.json"


@pytest.fixture
def mock_framework():
    """Mock the BalancingFramework class."""
    mock = MagicMock()
    # Set up common methods
    mock.inspect_class_distribution.return_value = {0: 75, 1: 25}
    return mock


@pytest.fixture
def mock_data_file(tmp_path):
    """Create a temporary CSV data file for testing."""
    data_file = tmp_path / "test_data.csv"
    with open(data_file, "w") as f:
        f.write("feature1,feature2,target\n")
        f.write("1,2,0\n")
        f.write("3,4,1\n")
    return data_file


@pytest.fixture
def args_load_data(mock_config_path, mock_data_file):
    """Create mock arguments for load_data command."""
    args = MagicMock()
    args.file_path = str(mock_data_file)
    args.target_column = "target"
    args.feature_columns = ["feature1", "feature2"]
    args.config_path = str(mock_config_path)
    args.verbose = False
    return args


@pytest.fixture
def args_preprocess(mock_config_path):
    """Create mock arguments for preprocess command."""
    args = MagicMock()
    args.handle_missing = "mean"
    args.scale = "standard"
    args.encode = "auto"
    args.config_path = str(mock_config_path)
    return args


class TestLoadDataCommand:
    """Tests for the load_data command."""

    @patch("balancr.cli.commands.BalancingFramework")
    @patch("balancr.cli.config.update_config")
    def test_load_data_success(
        self, mock_update_config, mock_framework_class, args_load_data, mock_framework
    ):
        """Test successful data loading."""
        mock_framework_class.return_value = mock_framework

        result = commands.load_data(args_load_data)

        # Verify framework was initialised and methods called
        mock_framework_class.assert_called_once()
        mock_framework.load_data.assert_called_once_with(
            args_load_data.file_path,
            args_load_data.target_column,
            args_load_data.feature_columns,
            correlation_threshold=0.95,
        )
        mock_framework.inspect_class_distribution.assert_called_once()

        # Verify config was updated
        mock_update_config.assert_called_once()
        settings = mock_update_config.call_args[0][1]
        assert settings["data_file"] == args_load_data.file_path
        assert settings["target_column"] == args_load_data.target_column
        assert settings["feature_columns"] == args_load_data.feature_columns

        # Verify result
        assert result == 0

    @patch("balancr.cli.commands.logging.error")
    def test_load_data_file_not_found(self, mock_error, args_load_data):
        """Test handling of non-existent data file."""
        # Modify args to point to non-existent file
        args_load_data.file_path = "nonexistent_file.csv"

        # Call function
        result = commands.load_data(args_load_data)

        # Verify error was logged and correct result returned
        mock_error.assert_called_once()
        assert "File not found" in mock_error.call_args[0][0]
        assert result == 1

    @patch("balancr.cli.commands.BalancingFramework")
    @patch("balancr.cli.commands.logging.error")
    def test_load_data_framework_error(
        self, mock_error, mock_framework_class, args_load_data, mock_framework
    ):
        """Test handling of framework error during data loading."""
        # Set up mock to raise an exception
        mock_framework_class.return_value = mock_framework
        mock_framework.load_data.side_effect = Exception("Test error")

        # Call function
        result = commands.load_data(args_load_data)

        # Verify error was logged and correct result returned
        mock_error.assert_called_once()
        assert "Failed to load data" in mock_error.call_args[0][0]
        assert result == 1

    @patch("balancr.cli.commands.BalancingFramework", None)
    @patch("balancr.cli.commands.logging.error")
    def test_load_data_no_framework(self, mock_error, args_load_data):
        """Test handling when BalancingFramework is not available."""
        # Call function
        result = commands.load_data(args_load_data)

        # Config should still be updated even without framework
        assert result == 0


class TestPreprocessCommand:
    """Tests for the preprocess command."""

    @patch("balancr.cli.config.load_config")
    @patch("balancr.cli.config.update_config")
    def test_preprocess_success(
        self, mock_update_config, mock_load_config, args_preprocess
    ):
        """Test successful preprocessing configuration."""
        # Set up load_config to return a valid config with data_file
        mock_load_config.return_value = {"data_file": "some_file.csv"}

        # Call the function
        result = commands.preprocess(args_preprocess)

        # Verify config was updated with correct settings
        mock_update_config.assert_called_once()
        settings = mock_update_config.call_args[0][1]
        assert (
            settings["preprocessing"]["handle_missing"]
            == args_preprocess.handle_missing
        )
        assert settings["preprocessing"]["scale"] == args_preprocess.scale
        assert settings["preprocessing"]["encode"] == args_preprocess.encode

        # Verify result
        assert result == 0

    @patch("balancr.cli.config.load_config")
    @patch("balancr.cli.commands.logging.error")
    def test_preprocess_no_data_file(
        self, mock_error, mock_load_config, args_preprocess
    ):
        """Test preprocessing without a configured data file."""
        # Set up load_config to return a config without data_file
        mock_load_config.return_value = {}

        # Call function
        result = commands.preprocess(args_preprocess)

        # Verify error was logged and function returned error code
        mock_error.assert_called_once()
        assert "Failed to configure preprocessing" in mock_error.call_args[0][0]
        assert result == 1

    @patch("balancr.cli.config.load_config")
    @patch("balancr.cli.config.update_config")
    @patch("balancr.cli.commands.logging.error")
    def test_preprocess_config_error(
        self, mock_error, mock_update_config, mock_load_config, args_preprocess
    ):
        """Test handling of configuration errors during preprocessing."""
        # Set up load_config to return a valid config
        mock_load_config.return_value = {"data_file": "some_file.csv"}

        # Make update_config raise an exception
        mock_update_config.side_effect = Exception("Config error")

        # Call function
        result = commands.preprocess(args_preprocess)

        # Verify error was logged and correct result returned
        assert "Error analysing dataset for encoding"
        assert (
            "Failed to configure preprocessing: Config error"
            in mock_error.call_args[0][0]
        )
        assert result == 1

    @patch("balancr.cli.config.load_config")
    @patch("builtins.print")
    @patch("balancr.cli.config.update_config")
    def test_preprocess_output(
        self, mock_update_config, mock_print, mock_load_config, args_preprocess
    ):
        """Test that preprocessing configuration outputs correct information."""
        # Set up load_config to return a valid config
        mock_load_config.return_value = {"data_file": "some_file.csv"}

        # Call function
        commands.preprocess(args_preprocess)

        # Verify the output contains the configuration details
        assert mock_print.call_count >= 3  # At least header and 3 settings

        # Verify the settings were printed (any order)
        settings_output = " ".join(
            str(call[0][0]) for call in mock_print.call_args_list
        )
        assert args_preprocess.handle_missing in settings_output
        assert args_preprocess.scale in settings_output
        assert args_preprocess.encode in settings_output
